@@ -2,14 +2,14 @@
 
 crate::prelude!();
 
-use rsmt2::{print::Expr2Smt, SmtConf};
+use rsmt2::print::Expr2Smt;
 
 use expr::{Expr, Var};
 use trans::Sys;
 
 pub mod cexs;
 
-pub use cexs::{Cexs, SmtParser, Solver};
+pub use cexs::Cexs;
 
 /// Aggregrates properties that are considered "ok" and properties that have been falsified.
 ///
@@ -59,7 +59,7 @@ impl<'s> CheckRes<'s> {
     ///
     /// - when `po` does not belong to `self.okay`;
     /// - when a falsification for `po` has already been registered.
-    pub fn register_falsification(&mut self, po: &'s String, solver: &mut Solver) -> Res<()> {
+    pub fn register_falsification(&mut self, po: &'s String, solver: &mut SFSolver) -> Res<()> {
         if self.cexs.contains_key(po) {
             bail!("trying to register PO `{}` as falsified twice")
         }
@@ -204,7 +204,7 @@ where
 /// they are not exposed outside of this crate.
 pub struct InternalChecker<'sys> {
     /// Underlying SMT solver.
-    solver: Solver,
+    solver: SFSolver,
     /// Transition system.
     sys: &'sys Sys,
     /// List of all variables of the system.
@@ -212,32 +212,8 @@ pub struct InternalChecker<'sys> {
 }
 impl<'sys> InternalChecker<'sys> {
     /// Constructor.
-    pub fn new(
-        sys: &'sys Sys,
-        z3_cmd: impl Into<String>,
-        tee: Option<impl AsRef<str>>,
-    ) -> Res<Self> {
-        let z3_cmd = z3_cmd.into();
-        let mut split_cmd = z3_cmd.split(|c: char| c.is_whitespace());
-        let z3_cmd = split_cmd
-            .next()
-            .ok_or_else(|| format!("illegal Z3 command `{}`", z3_cmd))?
-            .trim();
-        let mut conf = SmtConf::z3(z3_cmd);
-
-        for opt in split_cmd {
-            let opt = opt.trim();
-            if !opt.is_empty() {
-                conf.option(opt);
-            }
-        }
-
-        let mut solver = conf
-            .spawn(cexs::SmtParser)
-            .chain_err(|| "while spawning z3 solver")?;
-        if let Some(path) = tee {
-            solver.path_tee(path.as_ref())?
-        }
+    pub fn new(sys: &'sys Sys, conf: SmtConf, tee: Option<PathBuf>) -> Res<Self> {
+        let solver = SFSolver::new(conf, tee)?;
         let vars = sys.decls().all().collect();
         Ok(Self { solver, sys, vars })
     }
@@ -350,6 +326,11 @@ impl<'sys> InternalChecker<'sys> {
         let res = self.solver.check_sat()?;
         Ok(res)
     }
+
+    /// Accessor for the underlying solver.
+    pub fn solver(&mut self) -> &mut SFSolver {
+        &mut self.solver
+    }
 }
 
 /// Base (init) checker.
@@ -359,14 +340,13 @@ pub struct Base<'sys> {
 }
 impl<'sys> Base<'sys> {
     /// Constructor.
-    pub fn new(
-        sys: &'sys Sys,
-        z3_cmd: impl Into<String>,
-        tee: Option<impl AsRef<str>>,
-    ) -> Res<Self> {
-        let tee = tee.map(|s| format!("{}/base.smt2", s.as_ref()));
+    pub fn new(sys: &'sys Sys, conf: SmtConf, tee: Option<PathBuf>) -> Res<Self> {
+        let tee = tee.map(|mut path| {
+            path.push("base.smt2");
+            path
+        });
         Ok(Self {
-            checker: InternalChecker::new(sys, z3_cmd, tee)?,
+            checker: InternalChecker::new(sys, conf, tee)?,
         })
     }
 
@@ -387,14 +367,13 @@ pub struct Step<'sys> {
 }
 impl<'sys> Step<'sys> {
     /// Constructor.
-    pub fn new(
-        sys: &'sys Sys,
-        z3_cmd: impl Into<String>,
-        tee: Option<impl AsRef<str>>,
-    ) -> Res<Self> {
-        let tee = tee.map(|s| format!("{}/step.smt2", s.as_ref()));
+    pub fn new(sys: &'sys Sys, conf: SmtConf, tee: Option<PathBuf>) -> Res<Self> {
+        let tee = tee.map(|mut path| {
+            path.push("step.smt2");
+            path
+        });
         Ok(Self {
-            checker: InternalChecker::new(sys, z3_cmd, tee)?,
+            checker: InternalChecker::new(sys, conf, tee)?,
         })
     }
 
@@ -468,13 +447,16 @@ impl<'sys> Bmc<'sys> {
     /// Constructor.
     pub fn new(
         sys: &'sys Sys,
-        z3_cmd: impl Into<String>,
-        tee: Option<impl AsRef<str>>,
+        conf: SmtConf,
+        tee: Option<PathBuf>,
         res: BmcRes<'sys>,
     ) -> Res<Self> {
-        let tee = tee.map(|s| format!("{}/bmc.smt2", s.as_ref()));
+        let tee = tee.map(|mut path| {
+            path.push("bmc.smt2");
+            path
+        });
 
-        let mut checker = InternalChecker::new(sys, z3_cmd, tee)?;
+        let mut checker = InternalChecker::new(sys, conf, tee)?;
         checker.declare_vars(0)?;
         checker.assert_init()?;
         Ok(Self {
