@@ -1,21 +1,182 @@
 //! Backend parser (SMT-LIB) and frontend parser (requires the `parser` feature).
 
-crate::prelude!();
+prelude!();
 
 use expr::{Cst, Expr, Op, PExpr, SExpr, SVar, Typ, Var};
 use rsmt2::parse::IdentParser;
 use trans::Decls;
 
-#[cfg(feature = "parser")]
-use trans::Sys;
-
-mod ast;
 pub mod kw;
 
-#[cfg(all(test, feature = "parser"))]
+#[cfg(test)]
 mod test;
 
-pub use ast::*;
+/// A span in the input text.
+#[readonly::make]
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+pub struct Span {
+    /// Span's start (inclusive).
+    pub start: usize,
+    /// Span's end (exclusive).
+    pub end: usize,
+}
+impl Span {
+    /// Constructor.
+    pub fn new(start: usize, end: usize) -> Self {
+        debug_assert!(start <= end);
+        Span { start, end }
+    }
+    /// Merges two spans, `self`'s start and `other`'s end.
+    ///
+    /// - illegal if `self.start > other.end`.
+    pub fn merge(self, other: Self) -> Self {
+        (self.start, other.end).into()
+    }
+
+    /// Extracts the relevant line of the input, and the previous/next line if any.
+    pub fn pretty_of(self, text: &str) -> (Option<String>, usize, usize, String, Option<String>) {
+        if text.is_empty() {
+            assert_eq!(self.start, 0);
+            assert_eq!(self.end, 0);
+            return (None, 0, 0, "<EOI>".into(), None);
+        }
+        let mut lines = text.lines().enumerate();
+
+        let mut count = self.start;
+        let mut prev_line = None;
+
+        while let Some((row, line)) = lines.next() {
+            if line.len() >= count {
+                let (line, next) = {
+                    match lines.next().map(|(_, s)| s.to_string()) {
+                        Some(next) if next.is_empty() => (line.into(), None),
+                        Some(next) => (line.into(), Some(next)),
+                        None if text.ends_with('\n') => (line.into(), None),
+                        None => (format!("{}<EOI>", line), None),
+                    }
+                };
+                return (prev_line.map(String::from), row, count, line, next);
+            }
+
+            count -= line.len() + 1;
+            prev_line = Some(line);
+        }
+
+        panic!(
+            "illegal offset {} on text of length {}",
+            self.start,
+            text.len()
+        );
+    }
+
+    /// Pretty multi-line string representation.
+    pub fn pretty_ml_of(
+        self,
+        text: impl AsRef<str>,
+        style: impl Style,
+        msg: impl AsRef<str>,
+    ) -> String {
+        let text = text.as_ref();
+        let msg = msg.as_ref();
+        let (prev, row, col, line, next) = self.pretty_of(text);
+        let (row_str, col_str) = ((row + 1).to_string(), (col + 1).to_string());
+        let offset = {
+            let mut offset = 0;
+            let mut cnt = 0;
+            for c in line.chars() {
+                if cnt < col {
+                    offset += 1;
+                    cnt += c.len_utf8();
+                } else {
+                    break;
+                }
+            }
+            offset
+        };
+        let mut s = format!(
+            "at {}:{}\n{} |{}\n{} | {}",
+            style.bold(&row_str),
+            style.bold(&col_str),
+            " ".repeat(row_str.len()),
+            prev.as_ref()
+                .map(|s| format!(" {}", s))
+                .unwrap_or("".into()),
+            style.bold(&row_str),
+            line,
+        );
+        s.push_str(&format!(
+            "\n{} | {}{} {}",
+            " ".repeat(row_str.len()),
+            " ".repeat(offset),
+            style.red("^~~~"),
+            style.red(if msg.is_empty() { "here" } else { msg }),
+        ));
+        if let Some(next) = next {
+            s.push_str(&format!("\n{} | {}", " ".repeat(row_str.len()), next))
+        }
+        s
+    }
+}
+impl From<(usize, usize)> for Span {
+    fn from((start, end): (usize, usize)) -> Self {
+        Self::new(start, end)
+    }
+}
+
+/// Wraps something with a span.
+#[derive(Debug, Clone, Copy)]
+pub struct Spn<T> {
+    /// Value wrapped.
+    pub inner: T,
+    /// Span.
+    pub span: Span,
+}
+impl<T: PartialEq> PartialEq for Spn<T> {
+    fn eq(&self, that: &Self) -> bool {
+        self.inner == that.inner
+    }
+}
+impl<T: Eq> Eq for Spn<T> {}
+impl<T> Spn<T> {
+    /// Constructor.
+    pub fn new(inner: T, span: impl Into<Span>) -> Self {
+        let span = span.into();
+        Self { inner, span }
+    }
+
+    /// Applies an operation to the inner value.
+    pub fn map<U>(self, f: impl FnOnce(T) -> U) -> Spn<U> {
+        Spn {
+            inner: f(self.inner),
+            span: self.span,
+        }
+    }
+
+    /// Applies an operation yielding a result to the inner value.
+    pub fn res_map<U>(
+        self,
+        mut f: impl FnMut(T) -> Result<U, &'static str>,
+    ) -> Result<Spn<U>, &'static str> {
+        let inner = f(self.inner)?;
+        Ok(Spn::new(inner, self.span))
+    }
+
+    /// Unwraps `self`'s and `other`'s inner values and merges their spans.
+    pub fn unwrap_merge<U>(self, other: Spn<U>) -> (T, U, Span) {
+        (self.inner, other.inner, self.span.merge(other.span))
+    }
+}
+impl<T> Deref for Spn<T> {
+    type Target = T;
+    fn deref(&self) -> &T {
+        &self.inner
+    }
+}
+impl From<Spn<&str>> for Spn<String> {
+    fn from(spn: Spn<&str>) -> Self {
+        spn.map(|s| s.into())
+    }
+}
 
 /// Yields `true` if `ident` is a keyword.
 pub fn is_kw(ident: impl AsRef<str>) -> bool {
@@ -31,11 +192,11 @@ pub fn fail_if_kw(ident: impl AsRef<str>) -> Result<(), String> {
     }
 }
 
-#[cfg(feature = "parser")]
 peg::parser! {
     /// PEG parsing rules, requires the `parser` feature.
     ///
-    /// The [`DEMO` constant][crate::DEMO] illustrates and discusses the syntax expected by the parser.
+    /// The [`crate::TRANS_DEMO`] constant illustrates and discusses the syntax expected by the
+    /// parser.
     pub grammar rules() for str {
         /// Whitespace.
         pub rule whitespace() = quiet! {
@@ -65,7 +226,8 @@ peg::parser! {
             // Rust-style.
             "//"
             // Do not match inner/outer documentation.
-            [^ '/' | '!' | '\n' ]*
+            &[^ '/' | '!' ]
+            [^ '\n' ]*
             // Newline or EOI.
             ("\n" / ![_])
         }
@@ -168,6 +330,19 @@ peg::parser! {
         // Error message.
         / expected!("identifier")
 
+        /// Same as [`ident`], but allows dash (`-`) in non-leading position.
+        pub rule dash_ident() -> Spn<&'input str>
+        =
+            start:position!() id:$(
+                ident() ("-" ident())*
+            ) end:position!() {
+                Spn::new(id, (start, end))
+            }
+
+        /// Double-quoted string.
+        pub rule dbl_quoted() -> &'input str
+        = "\"" str:$( ("\\\"" / [^'"'])* ) "\"" { str }
+
 
         /// Parses boolean constants.
         ///
@@ -240,6 +415,25 @@ peg::parser! {
         }
         / expected!("integer")
 
+        /// Parses an isize.
+        pub rule isize() -> isize
+        = quiet! {
+            sign:("-" _)? code:number() {?
+                let neg = sign.is_some();
+                isize::from_str_radix(code.inner, 10)
+                    .map(|i|
+                        if neg {
+                            - i
+                        } else {
+                            i
+                        }
+                    ).map_err(|e|
+                        "illegal `isize` value"
+                    )
+            }
+        }
+        / expected!("isize value")
+
 
         /// Parses an unsigned [`Int`], cannot be followed by a `.`.
         ///
@@ -315,42 +509,42 @@ peg::parser! {
         / expected!("int/rat/bool constant")
 
         /// Parses variables.
-        pub rule hsmt_var() -> Ast<'input>
+        pub rule hsmt_var() -> ast::Expr<'input>
         = quiet! {
             prime:(
                 ps:position!() "'" pe:position!() {
                     Span::from((ps, pe))
                 }
             )? var:ident() {
-                Ast::svar(var, prime)
+                ast::Expr::svar(var, prime)
             }
         }
         / expected!("(un)primed state variable")
 
         /// Parses an if-then-else.
         ///
-        /// No parens needed, for documentation see [`hsmt`].
-        pub rule hsmt_ite() -> Ast<'input>
+        /// No parens needed, for documentation see [`hsmt_expr`].
+        pub rule hsmt_ite() -> ast::Expr<'input>
         = quiet! {
             s:position!() "if" e:position!()
-            _ cnd:hsmt()
+            _ cnd:hsmt_expr()
             _ "{"
-            _ thn:hsmt()
+            _ thn:hsmt_expr()
             _ "}"
             _ elseif:(
-                "else" _ s:position!() "if" e:position!() _ cnd:hsmt() _ "{" _ thn:hsmt() _ "}" {
+                "else" _ s:position!() "if" e:position!() _ cnd:hsmt_expr() _ "{" _ thn:hsmt_expr() _ "}" {
                     (Span::new(s, e), cnd, thn)
                 }
             )*
             _ "else" _ "{"
-            _ els:hsmt()
+            _ els:hsmt_expr()
             _ "}" {
                 let ite = Op::Ite;
                 let els = elseif.into_iter().rev().fold(
                     els,
-                    |els, (if_span, cnd, thn)| Ast::app(Spn::new(ite, if_span), vec![cnd, thn, els]),
+                    |els, (if_span, cnd, thn)| ast::Expr::app(Spn::new(ite, if_span), vec![cnd, thn, els]),
                 );
-                Ast::app(Spn::new(Op::Ite, (s,e)), vec![cnd, thn, els])
+                ast::Expr::app(Spn::new(Op::Ite, (s,e)), vec![cnd, thn, els])
             }
         }
         / expected!("if-then-else")
@@ -360,8 +554,8 @@ peg::parser! {
         /// # Examples
         ///
         /// ```rust
-        /// # use mikino_api::parse::rules::hsmt;
-        /// let ast = hsmt(
+        /// # use mikino_api::parse::rules::hsmt_expr;
+        /// let ast = hsmt_expr(
         ///     "if a ⋀ n ≥ 10 { 'n = n - 1 } \
         ///     else if false { 'n > n } \
         ///     else { false }"
@@ -379,8 +573,8 @@ peg::parser! {
         /// ```
         ///
         /// ```rust
-        /// use mikino_api::parse::rules::hsmt;
-        /// let ast = hsmt(
+        /// use mikino_api::parse::rules::hsmt_expr;
+        /// let ast = hsmt_expr(
         ///     "a ⋀ x ≥ 7 - m ⋁ if 7 = n + m { b_1 ⋁ b_2 } else { c }"
         /// ).unwrap();
         /// assert_eq!(
@@ -391,65 +585,65 @@ peg::parser! {
         ///     ",
         /// )
         /// ```
-        pub rule hsmt() -> Ast<'input>
+        pub rule hsmt_expr() -> ast::Expr<'input>
         = ast:precedence! {
             // Implication, right associative.
             lft:@ _ s:position!() (
                 "=>" / "⇒" / "→" / "⊃"
              ) e:position!() _ rgt:(@) {
-                Ast::binapp(Spn::new(Op::Implies, (s, e)), lft, rgt)
+                ast::Expr::binapp(Spn::new(Op::Implies, (s, e)), lft, rgt)
             }
             --
             lft:(@) _ s:position!() (
                 "∨" / "⋁" / "||" / "and"
              ) e:position!() _ rgt:@ {
-                Ast::binapp(Spn::new(Op::Or, (s, e)), lft, rgt)
+                ast::Expr::binapp(Spn::new(Op::Or, (s, e)), lft, rgt)
             }
             --
             lft:(@) _ s:position!() (
                 "∧" / "⋀" / "&&" / "or"
             ) e:position!() _ rgt:@ {
-                Ast::binapp(Spn::new(Op::And, (s, e)), lft, rgt)
+                ast::Expr::binapp(Spn::new(Op::And, (s, e)), lft, rgt)
             }
             --
             lft:(@) _ s:position!() "<" e:position!() _ rgt:@ {
-                Ast::binapp(Spn::new(Op::Lt, (s, e)), lft, rgt)
+                ast::Expr::binapp(Spn::new(Op::Lt, (s, e)), lft, rgt)
             }
             lft:(@) _ s:position!() ("<=" / "≤") e:position!() _ rgt:@ {
-                Ast::binapp(Spn::new(Op::Le, (s, e)), lft, rgt)
+                ast::Expr::binapp(Spn::new(Op::Le, (s, e)), lft, rgt)
             }
             lft:(@) _ s:position!() (">=" / "≥") e:position!() _ rgt:@ {
-                Ast::binapp(Spn::new(Op::Ge, (s, e)), lft, rgt)
+                ast::Expr::binapp(Spn::new(Op::Ge, (s, e)), lft, rgt)
             }
             lft:(@) _ s:position!() ">" e:position!() _ rgt:@ {
-                Ast::binapp(Spn::new(Op::Gt, (s, e)), lft, rgt)
+                ast::Expr::binapp(Spn::new(Op::Gt, (s, e)), lft, rgt)
             }
             lft:(@) _ s:position!() "=" e:position!() _ rgt:@ {
-                Ast::binapp(Spn::new(Op::Eq, (s, e)), lft, rgt)
+                ast::Expr::binapp(Spn::new(Op::Eq, (s, e)), lft, rgt)
             }
             --
             lft:(@) _ s:position!() "+" e:position!() _ rgt:@ {
-                Ast::binapp(Spn::new(Op::Add, (s, e)), lft, rgt)
+                ast::Expr::binapp(Spn::new(Op::Add, (s, e)), lft, rgt)
             }
             lft:(@) _ s:position!() "-" e:position!() _ rgt:@ {
-                Ast::binapp(Spn::new(Op::Sub, (s, e)), lft, rgt)
+                ast::Expr::binapp(Spn::new(Op::Sub, (s, e)), lft, rgt)
             }
             lft:(@) _ s:position!() "%" e:position!() _ rgt:@ {
-                Ast::binapp(Spn::new(Op::Mod, (s, e)), lft, rgt)
+                ast::Expr::binapp(Spn::new(Op::Mod, (s, e)), lft, rgt)
             }
             --
             lft:(@) _ s:position!() "*" e:position!() _ rgt:@ {
-                Ast::binapp(Spn::new(Op::Mul, (s, e)), lft, rgt)
+                ast::Expr::binapp(Spn::new(Op::Mul, (s, e)), lft, rgt)
             }
             lft:(@) _ s:position!() "/" e:position!() _ rgt:@ {
-                Ast::binapp(Spn::new(Op::Div, (s, e)), lft, rgt)
+                ast::Expr::binapp(Spn::new(Op::Div, (s, e)), lft, rgt)
             }
             --
             s:position!() ("¬" / "!" / "not") e:position!() _ arg:@ {
-                Ast::unapp(Spn::new(Op::Not, (s, e)), arg)
+                ast::Expr::unapp(Spn::new(Op::Not, (s, e)), arg)
             }
             s:position!() "-" e:position!() _ arg:@ {
-                Ast::unapp(Spn::new(Op::Sub, (s, e)), arg)
+                ast::Expr::unapp(Spn::new(Op::Sub, (s, e)), arg)
             }
             --
             ite:hsmt_ite() {
@@ -459,15 +653,21 @@ peg::parser! {
                 var
             }
             cst:cst() {
-                Ast::cst(cst)
+                ast::Expr::cst(cst)
             }
-            "(" _ e:hsmt() _ ")" {
+            "(" _ e:hsmt_expr() _ ")" {
                 let mut e = e;
                 e.close();
                 e
             }
         }
 
+        /// Same as [`hsmt_expr`].
+        pub rule hsmt_expr_with_repr() -> (ast::Expr<'input>, String)
+        = repr:&($(hsmt_expr())) expr:hsmt_expr() {
+            (expr, repr.into())
+        }
+
         /// Parses a type.
         ///
         /// Can be `int`, `rat`, or `bool`.
@@ -574,13 +774,11 @@ peg::parser! {
         /// assert_eq!(*name, "tautology");
         /// assert_eq!(expr.to_string(), "(p ⋁ (¬p))");
         /// ```
-        pub rule candidates() -> Vec<(Spn<&'input str>, Ast<'input>)>
+        pub rule candidates() -> Vec<(Spn<&'input str>, ast::Expr<'input>)>
         = quiet! {
             cands:(
-                _ s:position!() name:(
-                    "\"" name:$( ("\\\"" / [^'"'])* ) "\"" { name }
-                ) e:position!() _ ":" _
-                expr:hsmt()
+                _ s:position!() name:dbl_quoted() e:position!() _ ":" _
+                expr:hsmt_expr()
                 {
                     (Spn::new(name, (s, e)), expr)
                 }
@@ -592,8 +790,8 @@ peg::parser! {
 
         /// Parses a full instance.
         ///
-        /// Same documentation as [the `sys` function][super::sys].
-        pub rule hsmt_sys() -> PRes<trans::Sys>
+        /// Same documentation as [the `trans` function][crate::parse::trans].
+        pub rule hsmt_trans() -> PRes<trans::Sys>
         =
         sys_doc:inner_doc()
 
@@ -602,14 +800,14 @@ peg::parser! {
         init_doc:outer_doc()
         _ init_s:position!() "init" init_e:position!() _ "{" _ hsmt_init:(
             quiet! {
-                init:(hsmt()) ++ (_ "," _) (",")? { init }
+                init:(hsmt_expr()) ++ (_ "," _) (",")? { init }
             }
             / expected!("comma-separated list of stateless expressions")
          ) _  "}"
         trans_doc:outer_doc()
         _ trans_s:position!() "trans" trans_e:position!() _ "{" _ hsmt_trans:(
             quiet! {
-                trans:(hsmt()) ++ (_ "," _) (",")? { trans }
+                trans:(hsmt_expr()) ++ (_ "," _) (",")? { trans }
             }
             / expected!("comma-separated list of stateful expressions")
          ) _ "}"
@@ -617,8 +815,8 @@ peg::parser! {
         _ "candidates" _ "{" _ candidates:candidates() _ "}"
         _ {
             let decls = decls?;
-            let init = Ast::app(Spn::new(Op::And, (init_s, init_e)), hsmt_init).to_expr(&decls)?;
-            let trans = Ast::app(Spn::new(Op::And, (trans_s, trans_e)), hsmt_trans).to_sexpr(&decls)?;
+            let init = ast::Expr::app(Spn::new(Op::And, (init_s, init_e)), hsmt_init).to_expr(&decls)?;
+            let trans = ast::Expr::app(Spn::new(Op::And, (trans_s, trans_e)), hsmt_trans).to_sexpr(&decls)?;
 
             let mut pos = Map::new();
 
@@ -632,6 +830,275 @@ peg::parser! {
 
             Ok(trans::Sys::new(decls, init, trans, pos))
         }
+
+
+
+
+
+
+
+
+        /// A **non-empty** sequence of commands.
+        pub rule commands() -> PRes<ast::script::Commands<ast::Expr<'input>, ast::Expr<'input>>>
+        =
+            head:command() tail:(_ cmd:command() { cmd })* {
+                let mut cmds = Vec::with_capacity(tail.len());
+                cmds.push(head?);
+                for cmd in tail {
+                    cmds.push(cmd?);
+                }
+                Ok(cmds)
+            }
+
+        /// Command parser.
+        pub rule command() -> PRes<ast::script::Command<ast::Expr<'input>, ast::Expr<'input>>>
+        =
+            odoc:outer_doc() _ res:(
+                cmd:set_options() { Ok(cmd?.into()) }
+                /
+                cmd:const_decls() { Ok(cmd?.into()) }
+                /
+                cmd:mlet() { Ok(cmd?.into()) }
+                /
+                cmd:assert() { Ok(cmd?.into()) }
+                /
+                cmd:get_model() { Ok(cmd?.into()) }
+                /
+                cmd:get_values() { Ok(cmd?.into()) }
+                /
+                cmd:echo() { Ok(cmd?.into()) }
+                /
+                cmd:reset() { Ok(cmd?.into()) }
+                /
+                query:query() { Ok(query?.into()) }
+            ) {
+                res
+            }
+
+        /// Query parser.
+        pub rule query() -> PRes<ast::script::Query<ast::Expr<'input>, ast::Expr<'input>>>
+        =
+            q:block() { Ok(q?.into()) }
+            /
+            q:check_sat() { Ok(q?.into()) }
+            /
+            q:ite() { Ok(q?.into()) }
+            /
+            q:panic() { Ok(q?.into()) }
+            /
+            q:exit() { Ok(q?.into()) }
+
+
+
+
+
+
+        /// Block parser.
+        pub rule block() -> PRes<ast::script::Block<ast::Expr<'input>, ast::Expr<'input>>>
+        =
+            "{" commands:(_ cmd:command() { cmd })* _ "}" {
+                let mut content = Vec::with_capacity(commands.len());
+                for command in commands {
+                    content.push(command?)
+                }
+                Ok(ast::script::Block::new(content))
+            }
+
+        /// Check sat.
+        pub rule check_sat() -> PRes<ast::script::CheckSat>
+        =
+            start:position!() "check_sat" "!"? end:position!()
+            _ "("
+                // (_ key:$(ident()) _ ":" _ ""
+            _ ")" {
+                Ok(ast::script::CheckSat::new((start, end), None, None))
+            }
+            /
+            start:position!() "check_sat" "!"? end:position!()
+            _ "{"
+                // (_ key:$(ident()) _ ":" _ ""
+            _ "}" {
+                Ok(ast::script::CheckSat::new((start, end), None, None))
+            }
+
+        /// Ite.
+        pub rule ite() -> PRes<ast::script::Ite<ast::Expr<'input>, ast::Expr<'input>>>
+        =
+            start:position!() "if" end:position!()
+            _ cnd:(
+                cnd:check_sat() { cnd.map(Either::Right) }
+                /
+                cnd:ident() {
+                    let cnd = cnd.map(expr::MetaVar::new);
+                    Ok(Either::Left(cnd))
+                }
+            ) _ thn:block()
+            tail:(
+                // Else branch
+                _ "else" _ els:block()
+                // Otherwise branch (timeout/unknown)
+                otw:(
+                    _ "otherwise" _ otw:block() { otw }
+                )? {
+                    (els, otw)
+                }
+            )? {
+                let (els, otw) = tail.unwrap_or_else(
+                    || (Ok(ast::script::Block::new(vec![])), None)
+                );
+                let otw = if let Some(otw) = otw {
+                    Some(otw?)
+                } else {
+                    None
+                };
+                Ok(ast::script::Ite::new((start, end), cnd?, thn?, els?, otw))
+            }
+
+        /// A panic.
+        pub rule panic() -> PRes<ast::script::Panic>
+        =
+            start:position!() "panic" "!"? end:position!()
+            _ "{" _ msg:dbl_quoted() _ "}" {
+                Ok(ast::script::Panic::new((start, end), msg))
+            }
+            /
+            start:position!() "panic" "!"? end:position!()
+            _ "(" _ msg:dbl_quoted() _ ")" {
+                Ok(ast::script::Panic::new((start, end), msg))
+            }
+
+        /// An exit.
+        pub rule exit() -> PRes<ast::script::Exit>
+        =
+            start:position!() "exit" "!"? end:position!()
+            _ "(" _ code:(isize())? _ ")" {
+                Ok(ast::script::Exit::new((start, end), code))
+            }
+
+
+
+
+        /// A `set-option` key/value pair.
+        pub rule set_option() -> ast::script::SetOption
+        =
+            key:dash_ident() _ ":"
+            _ start:position!() val:(
+                cst:cst() { Either::Left(cst.inner) }
+                /
+                s:dbl_quoted() { Either::Right(s.to_string()) }
+            ) end:position!() {
+                ast::script::SetOption::new(key, Spn::new(val, (start, end)))
+            }
+
+        /// A sequence of `set-option`s.
+        pub rule set_options() -> PRes<ast::script::SetOptions>
+        =
+            start:position!() (
+                "set_options" / "set_option"
+            ) end:position!()
+            _ "("
+                opts:(
+                    _ opt:set_option() _ { opt }
+                )++"," _ ","?
+            _ ")" {
+                Ok(ast::script::SetOptions::new((start, end), opts))
+            }
+
+        /// A `declare-const` let-binding.
+        pub rule const_decls() -> PRes<ast::script::Vars>
+        =
+            start:position!() "vars" end:position!()
+            _ "{" _ decls:svars() _ "}" {
+                Ok(ast::script::Vars::new((start, end), decls?))
+            }
+            / start:position!() "vars" end:position!()
+            _ "(" _ decls:svars() _ ")" {
+                Ok(ast::script::Vars::new((start, end), decls?))
+            }
+
+        /// A meta-binding.
+        pub rule mlet() -> PRes<ast::script::MLet>
+        =
+            "let" _ lhs:ident() _ "=" _ query:check_sat() _ ";" {
+                Ok(ast::script::MLet::new(lhs, query?))
+            }
+
+        /// An assert.
+        pub rule assert() -> PRes<ast::script::Assert<ast::Expr<'input>>>
+        =
+            start:position!() "assert" end:position!()
+            _ "{" exprs:(_ expr:hsmt_expr() _ { expr })++"," _ ","? _ "}" {
+                Ok(ast::script::Assert::new((start, end), exprs))
+            }
+            /
+            start:position!() "assert" end:position!()
+            _ "(" exprs:(_ expr:hsmt_expr() _ { expr })++"," _ ","? _ ")" {
+                Ok(ast::script::Assert::new((start, end), exprs))
+            }
+
+        /// An assert.
+        pub rule get_model() -> PRes<ast::script::GetModel>
+        =
+            start:position!() token:$("get_model") "!"? end:position!() _ "(" _ ")" {
+                Ok(ast::script::GetModel::new((start, end), token))
+            }
+            /
+            start:position!() token:$("get_model") "!"? end:position!() _ "{" _ "}" {
+                Ok(ast::script::GetModel::new((start, end), token))
+            }
+
+        /// An assert.
+        pub rule get_values() -> PRes<ast::script::GetValues<ast::Expr<'input>>>
+        =
+            start:position!() token:$("get_value" "s"?/"eval") "!"? end:position!() _ "("
+                _ exprs:(_ expr:hsmt_expr_with_repr() _ { expr })++"," _ ","?
+            _ ")" {
+                Ok(ast::script::GetValues::new((start, end), token, exprs))
+            }
+            /
+            start:position!() token:$("get_values"/"eval") "!"? end:position!() _ "{"
+                _ exprs:(_ expr:hsmt_expr_with_repr() _ { expr })++"," _ ","?
+            _ "}" {
+                Ok(ast::script::GetValues::new((start, end), token, exprs))
+            }
+
+        /// An assert.
+        pub rule reset() -> PRes<ast::script::Reset>
+        =
+            start:position!() "reset" "!"? end:position!() _ "(" _ ")" {
+                Ok(ast::script::Reset::new((start, end)))
+            }
+            /
+            start:position!() "reset" "!"? end:position!() _ "{" _ "}" {
+                Ok(ast::script::Reset::new((start, end)))
+            }
+
+        /// An echo.
+        pub rule echo() -> PRes<ast::script::Echo>
+        =
+            start:position!()
+                token:$("echo"/"println") "!"?
+            end:position!() _ "{"
+                _ msg:dbl_quoted()? _
+            "}" {
+                Ok(ast::script::Echo::new((start, end), token, msg))
+            }
+            /
+            start:position!()
+                token:$("echo"/"println") "!"?
+            end:position!() _ "("
+                _ msg:dbl_quoted()? _
+            ")" {
+                Ok(ast::script::Echo::new((start, end), token, msg))
+            }
+
+        /// Parses a hsmt script.
+        pub rule hsmt_script() -> PRes<ast::script::Block<ast::Expr<'input>, ast::Expr<'input>>>
+        =
+            _ odoc:inner_doc()
+            _ content:commands() _ {
+                Ok(ast::script::Block::new(content?))
+            }
     }
 }
 
@@ -648,21 +1115,38 @@ peg::parser! {
 /// - `trans { ... }`: the transition relation, *i.e.* a stateful (`'` primes allowed) expression;
 ///
 /// - `candidates { ... }`: some [candidates][rules::candidates] to prove over the systems.
-#[cfg(feature = "parser")]
-pub fn sys(txt: &str) -> Res<Sys> {
-    let res: Res<Sys> = match rules::hsmt_sys(txt) {
-        Ok(res) => match res {
-            Ok(sys) => Ok(sys),
-            Err(e) => {
-                println!("perror");
-                let span = e.span;
-                let (prev, row, col, line, next) = span.pretty_of(txt);
-                let err = Error::parse("", row, col, line, prev, next);
-                Err(err.extend(e.error.into_iter()))
-            }
-        },
+pub fn trans(txt: &str) -> Res<trans::Sys> {
+    let res: Res<trans::Sys> = match rules::hsmt_trans(txt) {
+        Ok(res) => res.map_err(|e| e.into_error(txt)),
         Err(e) => {
-            println!("peg parse error");
+            // println!("peg parse error");
+            let span = Span::new(e.location.offset, e.location.offset);
+            let (prev, row, col, line, next) = span.pretty_of(txt);
+            let err = Error::parse("", row, col, line, prev, next);
+            Err(err.chain_err(|| format!("expected {}", e.expected)))
+        }
+    };
+    res.chain_err(|| "run mikino in 'demo' mode for more details about the syntax")
+}
+
+/// Parses a system, requires the `parser` feature.
+///
+/// Comments are one-line rust-style: `// ..\n`.
+///
+/// A system is composed of four `{ ... }` blocks each starting with a specific keyword:
+///
+/// - `svars { ... }`: the [state variables][rules::svars] of the system;
+///
+/// - `init { ... }`: the initial predicate, *i.e.* a stateless (no `'` prime) expression;
+///
+/// - `trans { ... }`: the transition relation, *i.e.* a stateful (`'` primes allowed) expression;
+///
+/// - `candidates { ... }`: some [candidates][rules::candidates] to prove over the systems.
+pub fn script(txt: &str) -> Res<ast::script::Block<ast::Expr, ast::Expr>> {
+    let res: Res<_> = match rules::hsmt_script(txt) {
+        Ok(res) => res.map_err(|e| e.into_error(txt)),
+        Err(e) => {
+            // println!("peg parse error");
             let span = Span::new(e.location.offset, e.location.offset);
             let (prev, row, col, line, next) = span.pretty_of(txt);
             let err = Error::parse("", row, col, line, prev, next);
